/*
 *  Driver for sampling hardware
 *  (c) 13.12.2014 Ingo D. Rullhusen
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 *
 */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/platform_device.h>
#include <linux/io.h>

#include <sound/core.h>
#include <sound/pcm.h>
#include <sound/initval.h>

#include <linux/device.h>


#include <linux/spinlock.h>
static DEFINE_SPINLOCK(slock);


#include <linux/interrupt.h>
#include <linux/gpio.h>

  
/* gpio base address */
static void __iomem *gpio_addr;

#define _PIO_BASE_ADDRESS (0x01c20800)

#define _Pn_CFG0(n) ( (n)*0x24 + 0x00 + gpio_addr )
#define _Pn_CFG1(n) ( (n)*0x24 + 0x04 + gpio_addr )
#define _Pn_CFG2(n) ( (n)*0x24 + 0x08 + gpio_addr )
#define _Pn_DRV1(n) ( (n)*0x24 + 0x18 + gpio_addr )
#define _Pn_PUL0(n) ( (n)*0x24 + 0x1C + gpio_addr )
#define _Pn_PUL1(n) ( (n)*0x24 + 0x20 + gpio_addr )



#define GPIO_DEVICE_DESC "DCsampler"

// PI14 -> C15 -> EINT26 -> GPIO-2 30

// PWM1/PI3 -> B19 -> GPIO-2 15
#define GPIO_PWM1      18  // global pin no.
#define GPIO_PWM1_DESC "clock"



#include "pwm.h"


#include "spi.h"




MODULE_AUTHOR("idr");
MODULE_DESCRIPTION("dcsampler soundcard");
MODULE_LICENSE("GPL");
MODULE_SUPPORTED_DEVICE("{{ALSA,dcsampler soundcard}}");


#define PERIOD_BYTES 384000
#define PERIODS 2
#define MAX_BUFFER (PERIOD_BYTES*PERIODS)


struct dcsampler_device
{
   struct snd_card *card;

   struct snd_pcm_substream *substream;


   unsigned int pcm_buffer_size; // in bytes


   struct device *spi_dev;
   sunxi_spi spi_data;


   sunxi_pwm pwm_data;

};




static struct snd_pcm_hardware dcsampler_pcm_hw =
{
        .info = ( SNDRV_PCM_INFO_INTERLEAVED |
                  SNDRV_PCM_INFO_BLOCK_TRANSFER ),
        .formats          = SNDRV_PCM_FMTBIT_S16_LE,
        .rates            = SNDRV_PCM_RATE_8000|SNDRV_PCM_RATE_KNOT,
        .rate_min         = 8000,
        .rate_max         = 352800,
        .channels_min     = 2,
        .channels_max     = 2,
        .buffer_bytes_max = MAX_BUFFER,
        .period_bytes_min = PERIOD_BYTES,
        .period_bytes_max = PERIOD_BYTES,
        .periods_min      = PERIODS,
        .periods_max      = PERIODS
};


static int dcsampler_pcm_open(struct snd_pcm_substream *ss)
{
   u32 value;
   unsigned long flags;
   int err = 0;
   struct dcsampler_device *mydev = ss->private_data;

   printk(KERN_INFO "dcsampler PCM open\n");

   mydev->substream = ss;

   ss->runtime->hw = dcsampler_pcm_hw;
   ss->runtime->private_data = mydev;

   spin_lock_irqsave(&slock, flags);


/*** init dma ***/
   if ( (err = spi_pcm_preallocate_dma_buffer( &mydev->spi_data, ss )) )
   { printk(KERN_ERR "prealloc error\n" ); return(-1); }



   value = readl( mydev->spi_data.ptr_ctl );
   value |= SPI_CTL_DMAMOD;
   value |= SPI_CTL_RST_RXFIFO | SPI_CTL_RST_TXFIFO; // reset
   value &= ~SPI_CTL_FUNC_MODE; // slave
   value |= SPI_CTL_EN; // enable
   writel( value, mydev->spi_data.ptr_ctl );

   value = readl( mydev->spi_data.ptr_dmactl );
   value |= SPI_DRQEN_RHF | SPI_DRQEN_RF;
   writel( value, mydev->spi_data.ptr_dmactl );


   spin_unlock_irqrestore(&slock, flags);


   return(err);
}

static int dcsampler_pcm_close(struct snd_pcm_substream *ss)
{
   u32 value;
   unsigned long flags;
   struct dcsampler_device *mydev = ss->private_data;

   printk(KERN_INFO "dcsampler PCM close\n");

   spin_lock_irqsave(&slock, flags);


/*** close dma ***/
   value = readl( mydev->spi_data.ptr_dmactl );
   value &= ~(SPI_DRQEN_RHF | SPI_DRQEN_RF);
   writel( value, mydev->spi_data.ptr_dmactl );



   spi_pcm_free_dma_buffers( &mydev->spi_data, ss );



   value = readl( mydev->spi_data.ptr_ctl );
   value &= ~SPI_CTL_EN; // disable
   writel( value, mydev->spi_data.ptr_ctl );

   spin_unlock_irqrestore(&slock, flags);



   ss->runtime->private_data = NULL;

   return(0);
}

static int dcsampler_hw_params(struct snd_pcm_substream *ss,
                               struct snd_pcm_hw_params *hw_params)
{
   return( snd_pcm_lib_malloc_pages(ss,params_buffer_bytes(hw_params)) );
}
 
static int dcsampler_hw_free(struct snd_pcm_substream *ss)
{
   return( snd_pcm_lib_free_pages(ss) );
}

static int dcsampler_pcm_prepare(struct snd_pcm_substream *ss)
{
   unsigned int bps;
   struct dcsampler_device *mydev = snd_pcm_substream_chip(ss);
   struct snd_pcm_runtime *runtime = ss->runtime;

printk(KERN_INFO "prepare\n");


   bps = runtime->rate * runtime->channels; // params requested by user ap
   bps *= snd_pcm_format_width(runtime->format);
   bps /= 8;
   if (bps <= 0)
      return -EINVAL;




   mydev->spi_data.buf_pos = 0;  
   mydev->pcm_buffer_size = frames_to_bytes(runtime, runtime->buffer_size);


   if (ss->stream == SNDRV_PCM_STREAM_CAPTURE)
   {
             /* clear capture buffer */
   }


/* set PWM rate */
   set_pwm1_rate( &mydev->pwm_data, runtime->rate );   


   return(0);
}

static int dcsampler_pcm_trigger(struct snd_pcm_substream *ss,
                                 int cmd)
{
   int err = 0;
   struct dcsampler_device *mydev = snd_pcm_substream_chip(ss);
   struct snd_pcm_runtime *runtime = ss->runtime;

   switch (cmd)
   {
      case SNDRV_PCM_TRIGGER_START:
      {
// Start the hardware capture
printk(KERN_INFO "trigger start\n");
printk(KERN_INFO "rate: %u\n", runtime->rate );

         enable_pwm1(&mydev->pwm_data);

         break;
      }
      case SNDRV_PCM_TRIGGER_STOP:
      {
// Stop the hardware capture


         disable_pwm1(&mydev->pwm_data);

         break;
      }
      default:
      {
printk(KERN_INFO "trigger unknown\n");

         err = -EINVAL;
      }
   }

   return(err);
}


static snd_pcm_uframes_t dcsampler_pcm_pointer(struct snd_pcm_substream *ss)
{
   struct dcsampler_device *mydev = snd_pcm_substream_chip(ss);
   struct snd_pcm_runtime *runtime = mydev->substream->runtime;

   return( bytes_to_frames(runtime,mydev->spi_data.buf_pos) );
}

typedef union { u32 ui; unsigned char b[4]; } frame;

static int dcsampler_pcm_copy(struct snd_pcm_substream *ss,
                              int channel, snd_pcm_uframes_t pos,
                              void __user *dst,
                              snd_pcm_uframes_t count)
{
   frame fr_t, *fptr1, *fptr2;
   snd_pcm_uframes_t fno;
   struct dcsampler_device *mydev = snd_pcm_substream_chip(ss);
   sunxi_spi *aw_spi = &mydev->spi_data;

///   struct snd_dma_buffer *buf = &ss->dma_buffer;
///   struct snd_pcm_runtime *runtime = mydev->substream->runtime;
///printk(KERN_INFO "pcm copy ch %i cn %i\n", channel, count );
///printk(KERN_INFO "pos %lu\n" , frames_to_bytes(ss->runtime,pos) ); // src pos
///printk(KERN_INFO "bufpos %lu\n" , aw_spi->buf_pos );

   fptr1 = (frame*)(aw_spi->snd_DMA->area) + pos;
   fptr2 = dst;
   for ( fno=0; fno<count; ++fno )
   {
      fr_t.b[0] = fptr1->b[1];
      fr_t.b[1] = fptr1->b[0];
      fr_t.b[2] = fptr1->b[3];
      fr_t.b[3] = fptr1->b[2];

      fptr2->ui = (fr_t.ui ^ 0x80008000);

      ++fptr1;
      ++fptr2;
   }


   return(0);
}



static struct snd_pcm_ops dcsampler_pcm_ops =
{
   .open      = dcsampler_pcm_open,
   .close     = dcsampler_pcm_close,
   .ioctl     = snd_pcm_lib_ioctl,
   .hw_params = dcsampler_hw_params,
   .hw_free   = dcsampler_hw_free, 
   .prepare   = dcsampler_pcm_prepare,
   .trigger   = dcsampler_pcm_trigger,
   .pointer   = dcsampler_pcm_pointer,
   .copy      = dcsampler_pcm_copy
};

static struct snd_device_ops dev_ops =
{
   NULL
};



/* device (one only) */
static struct platform_device *device;

#define SND_DCSAMPLER_DRIVER "snd_dcsampler"



#include "spictl.i"

static struct class_attribute spictl_class_attrs[] =
{ __ATTR_NULL };
 
static struct class spictl_class =
{
   .name =         "snd-spi",
   .owner =        THIS_MODULE,
   .class_attrs =  spictl_class_attrs,
};



static int __devinit dcsampler_probe(struct platform_device *devptr)
{
   int err = 0;
   u32 value;
   unsigned long nv, flags;

   struct dcsampler_device *mydev;
   struct snd_card *card;

   struct snd_pcm *pcm;

   printk( KERN_INFO "probe dev dcsampler\n" );

/* one card only */
   err = snd_card_create( SNDRV_DEFAULT_IDX1, "DCsampler",
                          THIS_MODULE, sizeof(struct dcsampler_device), &card );

   if ( err )
   {
      printk( KERN_ERR "dcsampler card create error\n" );
      return(err);
   }

   mydev = card->private_data;
   mydev->card = card;

   strcpy( card->driver, "DCsampler driver" );
   strcpy( card->shortname, "DC sampler" );
   sprintf( card->longname, "%s on SPI", card->shortname );

   snd_card_set_dev(card, &devptr->dev);

   if ( (err = snd_device_new(card, SNDRV_DEV_LOWLEVEL, mydev, &dev_ops)) )
   {
      printk( KERN_ERR "dcsampler dev new error\n" );
      return(err);
   }


/* PCM */

   if ( (err = snd_pcm_new(card, card->driver, 0, 0, 1, &pcm)) )
   {
      printk( KERN_ERR "dcsampler PCM new error\n" );
      return(err);
   }

   snd_pcm_set_ops(pcm, SNDRV_PCM_STREAM_CAPTURE, &dcsampler_pcm_ops);
   pcm->private_data = mydev;
   pcm->info_flags = 0;
   strcpy(pcm->name, "DCsampler PCM");

   err = snd_pcm_lib_preallocate_pages_for_all(pcm,
            SNDRV_DMA_TYPE_CONTINUOUS,
            snd_dma_continuous_data(GFP_KERNEL),
            MAX_BUFFER, MAX_BUFFER);

   if (err < 0)
   {
      printk( KERN_ERR "dcsampler PCM prealloc error\n" );
      return(err);
   }


   if ( (err = snd_card_register(card)) )
   {
      printk( KERN_ERR "dcsampler register error\n" );
      return(err);
   }
   else
   {
      platform_set_drvdata(devptr, card);
   }


/* IRQ */



/* SPI 1 init */
printk( KERN_INFO "vorher\n" );
   if ( (err = class_register(&spictl_class)) )
   { printk( KERN_ERR "kann nicht spi reggen!\n" ); return(-1); }

   mydev->spi_dev = device_create( &spictl_class, NULL,
                                   MKDEV(0, 0), &mydev->spi_data, spi_name[1] );
   if ( ! mydev->spi_dev )
   { printk( KERN_ERR "kann nicht spi dev!\n" ); return(-1); }

   if ( (err = sysfs_create_group(&mydev->spi_dev->kobj, &spi_attr_group)) )
   { printk( KERN_ERR "kann nicht syscreaten!\n" ); return(-1); }
      
   if ( (err=spi_data_init( mydev->spi_dev, &mydev->spi_data, 1 )) )
   { printk( KERN_ERR "kann nicht spi initen!\n" ); return(-1); }


   spin_lock_irqsave(&slock, flags);

   value = readl( mydev->spi_data.ptr_cctl );

      value &= ~SPI_CLKCTL_DRS; // false

      nv = 3;
      value &= ~SPI_CLKCTL_CDR1;
      value |= (SPI_CLKCTL_CDR1&(nv<<8));
   
   writel( value, mydev->spi_data.ptr_cctl );

   value = readl( mydev->spi_data.ptr_wait );
   
      nv = 2;
      value &= ~SPI_WAIT_CLK_MASK;
      value |= (SPI_WAIT_CLK_MASK&(nv<<0));

   writel( value, mydev->spi_data.ptr_wait );

   value = readl( mydev->spi_data.ptr_dmactl );

      value |= SPI_DRQEN_RHF; // true
      value |= SPI_DRQEN_RF;

   writel( value, mydev->spi_data.ptr_dmactl );

   value = readl( mydev->spi_data.ptr_intctl );

      value &= ~SPI_INTEN_RH;
      value &= ~SPI_INTEN_RO;

   writel( value, mydev->spi_data.ptr_intctl );

   value = readl( mydev->spi_data.ptr_ctl );

      value |= SPI_CTL_RST_RXFIFO;
      value |= SPI_CTL_RST_TXFIFO;

      value |= SPI_CTL_DMAMOD;

      value &= ~SPI_CTL_FUNC_MODE; // slave
      value &= ~SPI_CTL_EN;

   writel( value, mydev->spi_data.ptr_ctl );

   spin_unlock_irqrestore(&slock, flags);






/* PWM init */
   if ( pwm_data_init( &mydev->pwm_data ) )
   { printk( KERN_ERR "kann nicht pwm initen!\n" ); return(-1); }




   return(err);
}

static int __devexit dcsampler_remove(struct platform_device *devptr)
{
   struct snd_card *card = platform_get_drvdata(devptr);
   struct dcsampler_device *mydev = card->private_data;

   printk( KERN_INFO "remove dev dcsampler\n" );



/* SPI 1 free */
   spi_data_free( mydev->spi_dev, &mydev->spi_data, 1 );
   device_destroy(&spictl_class, mydev->spi_dev->devt);
   class_unregister(&spictl_class);

/* PWM free */
   pwm_data_free( &mydev->pwm_data );



   snd_card_free(card);
   platform_set_drvdata(devptr, NULL);

   return(0);
}

static struct platform_driver dcsampler_driver =
{
   .probe = dcsampler_probe,
   .remove = __devexit_p(dcsampler_remove),
   .driver = {
      .name  = SND_DCSAMPLER_DRIVER,
      .owner = THIS_MODULE
             }
};






static int __init dcsampler_init(void)
{
   int err = 0;
   unsigned int reg_val;
   unsigned long flags;

   printk( KERN_INFO "init dcsampler\n" );

/* mem & I/O spec */

   gpio_addr = ioremap( _PIO_BASE_ADDRESS, 0x400 );

   spin_lock_irqsave( &slock, flags );

/* UEXT2 */
   reg_val = readl(_Pn_CFG2(8));
   reg_val &= ~( 0x00007777 );
   reg_val |=  ( 0x00002222 );
   writel(reg_val, _Pn_CFG2(8) );

/* PWM1/PI3 */
   reg_val = readl(_Pn_CFG0(8));
   reg_val &= ~( 0x00007000 );
   reg_val |=  ( 0x00002000 );
   writel(reg_val, _Pn_CFG0(8) );

   reg_val = readl(_Pn_PUL0(8));
   reg_val &= ~( 0x000000c0 );
   reg_val |=  ( 0x00000000 ); // no pull
   writel(reg_val, _Pn_PUL0(8) );

   spin_unlock_irqrestore( &slock, flags );


/* driver/device calls _probe */

   if ( (err = platform_driver_register(&dcsampler_driver)) )
   {
      printk( KERN_ERR "dcsampler driver register error\n" );
      return(err);
   }

   device = platform_device_register_simple(SND_DCSAMPLER_DRIVER, 0, NULL, 0);

   if (IS_ERR(device))
   {
      printk(KERN_ERR "dcsampler: not enabled, not found or device busy\n");
      return(-ENODEV);
   }









   return(0);
}

static void __exit dcsampler_exit(void)
{
   unsigned int reg_val;
   unsigned long flags;

   printk( KERN_INFO "exit dcsampler\n" );




   platform_device_unregister(device);

   platform_driver_unregister(&dcsampler_driver);


   spin_lock_irqsave( &slock, flags );

/* PWM1/PI3 */
   reg_val = readl(_Pn_CFG0(8));
   reg_val &= ~( 0x00007000 );
   reg_val |=  ( 0x00000000 );
   writel(reg_val, _Pn_CFG0(8) );

/* UEXT2 */
   reg_val = readl(_Pn_CFG2(8));
   reg_val &= ~( 0x00007777 );
   reg_val |=  ( 0x00000000 );
   writel(reg_val, _Pn_CFG2(8) );

   spin_unlock_irqrestore( &slock, flags );


   iounmap(gpio_addr);

}

module_init(dcsampler_init);
module_exit(dcsampler_exit);
 
MODULE_LICENSE("GPL"); 
MODULE_AUTHOR("idr");

