/* spi.c
 *
 * (c) 23.12.2014 Ingo D. Rullhusen
 *
 */


#include <linux/kernel.h>
#include <linux/err.h>
#include <linux/ioport.h>
#include <linux/spinlock.h>
#include <linux/interrupt.h>
#include <linux/platform_device.h>
#include <linux/slab.h>
#include <linux/dma-mapping.h>

#include <linux/gpio.h>

#include <mach/dma.h>

#include <asm/io.h>
#include <asm/dma.h>

#include <sound/core.h>
#include <sound/pcm.h>


#include "spi.h"


static char *spi_clk_name[] = { "ahb_spi0", "ahb_spi1", "ahb_spi2", "ahb_spi3" };

static char *spi_dma_rx[] = {"spi0_rx", "spi1_rx", "spi2_rx", "spi3_rx"};
static char *spi_dma_tx[] = {"spi0_tx", "spi1_tx", "spi2_tx", "spi3_tx"};


static DEFINE_SPINLOCK(slock);



/* set spi clock */
void spi_set_clk(u32 spi_clk, u32 ahb_clk, void *base_addr)
{
    u32 reg_val = 0;
    u32 N = 0;
    u32 div_clk = (ahb_clk >> 1) / spi_clk;

    reg_val = readl(base_addr + SPI_CLK_RATE_REG);

    if (div_clk <= SPI_CLK_SCOPE) {
        if (div_clk != 0) {
            div_clk--;
        }

        reg_val &= ~SPI_CLKCTL_CDR2;
        reg_val |= (div_clk | SPI_CLKCTL_DRS);
    } else {
        while (1) {
            if (div_clk == 1) {
                break;
            }
            div_clk >>= 1;
            N++;
        };

        reg_val &= ~(SPI_CLKCTL_CDR1 | SPI_CLKCTL_DRS);
        reg_val |= (N << 8);
    }

    writel(reg_val, base_addr + SPI_CLK_RATE_REG);
}




/* disable dma irq */
void spi_disable_dma_irq(u32 bitmap, void *base_addr)
{
    u32 reg_val = readl(base_addr + SPI_DMA_CTL_REG);
    bitmap &= SPI_DRQEN_MASK;
    reg_val &= ~bitmap;
    writel(reg_val, base_addr + SPI_DMA_CTL_REG);
}
 
/* enable dma irq */
void spi_enable_dma_irq(u32 bitmap, void *base_addr)
{
    u32 reg_val = readl(base_addr + SPI_DMA_CTL_REG);
    bitmap &= SPI_DRQEN_MASK;
    reg_val |= bitmap;
    writel(reg_val, base_addr + SPI_DMA_CTL_REG);
}



static void sun7i_spi_dma_cb_rx_full(dma_hdl_t dma_hdl, void *parg)   
{
   sunxi_spi *aw_spi = parg;

///printk( KERN_INFO "DMA full...\n" );

if ( !aw_spi->snd_DMA )
{ printk( KERN_ERR "DMA not ready full\n" ); return; }

if ( aw_spi->buf_pos == aw_spi->snd_DMA->bytes )
{ printk( KERN_ERR "multicall full\n" ); }

   aw_spi->buf_pos = aw_spi->snd_DMA->bytes - 1;

   snd_pcm_period_elapsed(aw_spi->substream); // new data

}


static void sun7i_spi_dma_cb_rx_half(dma_hdl_t dma_hdl, void *parg)   
{
   sunxi_spi *aw_spi = parg;

///printk( KERN_INFO "DMA half...\n" );

if ( !aw_spi->snd_DMA )
{ printk( KERN_ERR "DMA not ready half\n" ); return; }

if ( aw_spi->buf_pos == aw_spi->snd_DMA->bytes/2 )
{ printk( KERN_ERR "multicall half\n" ); }

   aw_spi->buf_pos = aw_spi->snd_DMA->bytes/2 - 1;

   snd_pcm_period_elapsed(aw_spi->substream); // new data

}



static unsigned char spi_rx_ddrq[] = {D_SRC_SPI0_RX, D_SRC_SPI1_RX, D_SRC_SPI2_RX, D_SRC_SPI3_RX};
static unsigned char spi_tx_ddrq[] = {D_DST_SPI0_TX, D_DST_SPI1_TX, D_DST_SPI3_TX, D_DST_SPI3_TX};
static phys_addr_t paddr[] = { SPI0_BASE_ADDR, SPI1_BASE_ADDR, SPI2_BASE_ADDR, SPI3_BASE_ADDR };


int spi_data_init( struct device *dev, sunxi_spi *SPI_DATA, int i ) /* i: 0...3 */
{
   struct clk *source_clock = NULL;

   SPI_DATA->name = dev_name(dev);
   printk( KERN_INFO "INIT NAME %s\n", SPI_DATA->name );

   if ( i<0 || i>3 ) { printk(KERN_ERR "wrong SPI number\n"); return(-1); }

   SPI_DATA->paddr = paddr[i];

   SPI_DATA->snd_DMA = NULL;
   SPI_DATA->buf_pos = 0;
   SPI_DATA->substream = NULL;

   SPI_DATA->spi_dma_rx = spi_dma_rx[i];
   SPI_DATA->spi_dma_tx = spi_dma_tx[i];
   SPI_DATA->spi_rx_ddrq = spi_rx_ddrq[i];
   SPI_DATA->spi_tx_ddrq = spi_tx_ddrq[i];

   request_mem_region( SPI_DATA->paddr, 0x1000, SPI_DATA->name );
   SPI_DATA->base_ptr = ioremap(SPI_DATA->paddr,0x1000);

   SPI_DATA->ptr_rxdata = SPI_DATA->base_ptr + SPI_RXDATA_REG;
   SPI_DATA->ptr_txdata = SPI_DATA->base_ptr + SPI_TXDATA_REG;
   SPI_DATA->ptr_ctl = SPI_DATA->base_ptr + SPI_CTL_REG;
   SPI_DATA->ptr_intctl = SPI_DATA->base_ptr + SPI_INT_CTL_REG;
   SPI_DATA->ptr_int_sta = SPI_DATA->base_ptr + SPI_STATUS_REG;
   SPI_DATA->ptr_dmactl = SPI_DATA->base_ptr + SPI_DMA_CTL_REG;
   SPI_DATA->ptr_wait = SPI_DATA->base_ptr + SPI_WAIT_REG;
   SPI_DATA->ptr_cctl = SPI_DATA->base_ptr + SPI_CLK_RATE_REG;
   SPI_DATA->ptr_bc = SPI_DATA->base_ptr + SPI_BC_REG;
   SPI_DATA->ptr_tc = SPI_DATA->base_ptr + SPI_TC_REG;
   SPI_DATA->ptr_fifo_sta = SPI_DATA->base_ptr + SPI_FIFO_STA_REG;


   SPI_DATA->ahb_clock = clk_get( dev, spi_clk_name[i] );
   if ( IS_ERR(SPI_DATA->ahb_clock) ) {
       printk( KERN_ERR "%s: Unable to get spi%d AHB clock resource\n",
               __func__, i );
       return -1;
   }
   clk_enable(SPI_DATA->ahb_clock);


   SPI_DATA->clock = clk_get(NULL, SPI_DATA->name );
   if ( IS_ERR(SPI_DATA->clock) ) {
       printk( KERN_ERR "%s: Unable to get %s source clock resource\n",
               __func__, SPI_DATA->name );
       return -1;
   }
   clk_prepare(SPI_DATA->clock);

   source_clock = clk_get(NULL, "sdram_pll_p" );
   clk_prepare(source_clock);

   clk_set_parent( SPI_DATA->clock, source_clock );

   clk_set_rate( SPI_DATA->clock, 100000000 );

   if ( clk_enable(SPI_DATA->clock) )
   {
      printk( KERN_ERR "cannot enable clock %d\n",i );
   }

   clk_put(source_clock);


   spi_set_clk( 500000, 24000000, SPI_DATA->base_ptr );


   SPI_DATA->DMA_rx = NULL;
   SPI_DATA->DMA_tx = NULL;


   return(0);
}

void spi_data_free( struct device *dev, sunxi_spi *SPI_DATA, int i )
{

   if ( SPI_DATA->ahb_clock )
   {
   clk_disable(SPI_DATA->ahb_clock);
   clk_put(SPI_DATA->ahb_clock);
   }
   if ( SPI_DATA->clock )
   {
   clk_disable(SPI_DATA->clock);
   clk_put(SPI_DATA->clock);
   }


   iounmap(SPI_DATA->base_ptr);

   release_mem_region( SPI_DATA->paddr, 0x1000 );

}


int spi_pcm_preallocate_dma_buffer(sunxi_spi *SPI_DATA, struct snd_pcm_substream *substream)
{
   struct snd_dma_buffer *buf = &substream->dma_buffer;

   int ret = 0;
   u32 uret = 0;

   dma_cb_t done_cb;
   dma_config_t spi_hw_conf;

   if (!buf->area)
   {
      printk(KERN_ERR "DMA present error\n" );
      return -ENOMEM;
   }
   SPI_DATA->snd_DMA = buf;
   SPI_DATA->buf_pos = 0;
   SPI_DATA->substream = substream;

printk(KERN_ERR "dma buff size %u\n", SPI_DATA->snd_DMA->bytes );


   SPI_DATA->DMA_rx = sw_dma_request( SPI_DATA->spi_dma_rx, CHAN_DEDICATE );
   if ( ! SPI_DATA->DMA_rx )
   { printk( KERN_ERR "dma req rx failed" ); return(-1); }

   done_cb.func = sun7i_spi_dma_cb_rx_full;
   done_cb.parg = SPI_DATA;
   if ( (uret=sw_dma_ctl( SPI_DATA->DMA_rx, DMA_OP_SET_FD_CB, (void *)&done_cb )) )
   {
      printk( KERN_ERR "%s: %s set dma rx DMA_OP_SET_FD_CB failed: %u\n", __func__, SPI_DATA->name, uret );
      return -1;
   }

printk(KERN_INFO "bis hier in dma alloc\n");
   done_cb.func = sun7i_spi_dma_cb_rx_half;
   done_cb.parg = SPI_DATA;
   if ( (uret=sw_dma_ctl( SPI_DATA->DMA_rx, DMA_OP_SET_HD_CB, (void *)&done_cb )) )
   {
      printk( KERN_ERR "%s: %s set dma rx DMA_OP_SET_HD_CB failed: %u\n", __func__, SPI_DATA->name, uret );
      return -1;
   }

   spi_hw_conf.xfer_type.src_data_width    = DATA_WIDTH_32BIT;
   spi_hw_conf.xfer_type.src_bst_len       = DATA_BRST_8;
   spi_hw_conf.xfer_type.dst_data_width    = DATA_WIDTH_32BIT;
   spi_hw_conf.xfer_type.dst_bst_len       = DATA_BRST_8;
   spi_hw_conf.bconti_mode                 = true;
   spi_hw_conf.irq_spt                     = CHAN_IRQ_HD | CHAN_IRQ_FD;

   spi_hw_conf.address_type.src_addr_mode  = DDMA_ADDR_IO;
   spi_hw_conf.address_type.dst_addr_mode  = DDMA_ADDR_LINEAR;
   spi_hw_conf.src_drq_type                = SPI_DATA->spi_rx_ddrq;
   spi_hw_conf.dst_drq_type                = D_DST_SDRAM;     

   if (sw_dma_config(SPI_DATA->DMA_rx, &spi_hw_conf))
   {
      printk( KERN_ERR "%s: %s rx config failed\n", __func__, SPI_DATA->name );
      return(-1);
   }


printk(KERN_INFO "bis hier 2 in dma alloc\n");
        dma_para_t para = {
            .src_wait_cyc = 0x0,
            .src_blk_sz   = 0x07,
            .dst_wait_cyc = 0x0,
            .dst_blk_sz   = 0x07
        };

        if (sw_dma_ctl( SPI_DATA->DMA_rx, DMA_OP_SET_PARA_REG, &para)) {
            printk( KERN_ERR "%s: %s set dma rx para failed\n", __func__, SPI_DATA->name );
            return -1;
        }


printk(KERN_INFO "bis hier 3 in dma alloc\n");
   if ( sw_dma_enqueue( SPI_DATA->DMA_rx,
                        SPI_DATA->paddr+SPI_RXDATA_REG,
                        virt_to_phys(SPI_DATA->snd_DMA->area),
                        SPI_DATA->snd_DMA->bytes ) )
   {
      printk( KERN_ERR "%s: %s set dma rx enqueue failed\n", __func__, SPI_DATA->name );
      return(-1);
   }


ret = sw_dma_ctl( SPI_DATA->DMA_rx, DMA_OP_START, NULL);
if ( ret )
{
   printk( KERN_ERR "DMA start failed\n" );
}




   return 0;
}

void spi_pcm_free_dma_buffers(sunxi_spi *SPI_DATA, struct snd_pcm_substream *substream)
{
   SPI_DATA->snd_DMA = NULL;
   SPI_DATA->buf_pos = 0;
   SPI_DATA->substream = NULL;

   sw_dma_ctl( SPI_DATA->DMA_rx, DMA_OP_STOP, NULL );
   sw_dma_release( SPI_DATA->DMA_rx );

}
